#include "codes.h"
#include "classes.h"

CCpu::CCpu(CMemory *pmArg,UI32 u32Arg) {

	pmMemory=pmArg;

	GR_LAST_INDEX=7;
	REG_SP=0x0d;
	REG_SR=0x0e;
	REG_PC=0x0f;

	SR_ZERO=0x00000001;
	SR_SIG= 0x00000002;
	SR_OF=  0x00000004;

	reset();

	setPC(u32Arg);

}

bool CCpu::reset() {

	for(int i=0;i<16;i++)
		u32Reg[i]=0;

	return true;

};

UI32 CCpu::getRegister(I32 i32Arg) const {

	if (i32Arg<0 || i32Arg>GR_LAST_INDEX)
		throw out_of_range("ȃWX^w");

	return u32Reg[i32Arg];

}

bool CCpu::setRegister(I32 i32Arg,UI32 u32Arg) {

	if (i32Arg<0 || i32Arg>GR_LAST_INDEX)
		throw out_of_range("ȃWX^w");

	u32Reg[i32Arg]=u32Arg;

	return true;

}

UI32 CCpu::getPC() const { return u32Reg[REG_PC]; }
UI32 CCpu::getSP() const { return u32Reg[REG_SP]; }
UI32 CCpu::getSR() const { return u32Reg[REG_SR]; }

bool CCpu::setPC(UI32 u32Arg) {

	u32Reg[REG_PC]=u32Arg;

	return true;

}

bool CCpu::setSP(UI32 u32Arg) {

	u32Reg[REG_SP]=u32Arg;

	return true;

}

bool CCpu::setSR(UI32 u32Arg) {

	u32Reg[REG_SR]=u32Arg;

	return true;

}

bool CCpu::getSRZero() const { return (u32Reg[REG_SR] & SR_ZERO!=0); }
bool CCpu::getSRSig() const { return (u32Reg[REG_SR] & SR_SIG!=0); }
bool CCpu::getSROF() const { return (u32Reg[REG_SR] & SR_OF!=0); }

bool CCpu::setSRFlag(UI32 u32Arg,bool bArg) {
	
	if (bArg) {

		setSR(getSR() | u32Arg);

		return true;

	} else {

		setSR(getSR() & ~u32Arg);

		return true;

	}

}

bool CCpu::setSROF(bool bArg) {

	return setSRFlag(SR_OF,bArg);

}

bool CCpu::setSRZero(bool bArg) {

	return setSRFlag(SR_ZERO,bArg);

}

bool CCpu::setSRSig(bool bArg) {

	return setSRFlag(SR_SIG,bArg);

}

bool CCpu::clearSR() {

	setSR(0);

	return true;

}

bool CCpu::exec() throw(logic_error) {

	exec(getPC());

	return true;

}

bool CCpu::u32OF(UI32 u32Arg1,UI32 u32Arg2) const {

	UI32 u32Wrk=0xffffffff;

	u32Wrk-=u32Arg1;

	return u32Wrk<u32Arg2;

}

bool CCpu::u32MulOF(UI32 u32Arg1,UI32 u32Arg2) const {

	UI32 u32Wrk=0xffffffff;

	u32Wrk/=u32Arg2;

	return u32Arg1>u32Wrk;

}

bool CCpu::exec(UI32 u32Arg) {

	UI32 u32Code=pmMemory->read32(u32Arg);
	UI32 u32Wrk,u32Wrk1,u32Wrk2,u32Wrk3,u32Addr;
	UI16 u16Wrk1,u16Wrk2;
	UI8 u8Wrk1,u8Wrk2;

	switch ((u32Code & 0xffff0000) >> 16) {

	case LDRR:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u32Wrk2=u32Code & 0x00ff;

		setRegister(u32Wrk1,getRegister(u32Wrk2));

		u32Addr=u32Arg+4;

		break;

	case LDRI8:

		u32Wrk1=(u32Code & 0xff00) >> 8;

		setRegister(u32Wrk1,u32Code & 0x00ff);

		u32Addr=u32Arg+4;

		break;

	case LDRI16:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u32Wrk2=pmMemory->read16(u32Arg+4);

		setRegister(u32Wrk1,u32Wrk2);

		u32Addr=u32Arg+6;

		break;

	case LDRI32:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u32Wrk2=pmMemory->read32(u32Arg+4);

		setRegister(u32Wrk1,u32Wrk2);

		u32Addr=u32Arg+8;

		break;

	case LDRM8:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u8Wrk1=pmMemory->read8(getRegister(u32Code & 0xff));

		setRegister(u32Wrk1,u8Wrk1);

		u32Addr=u32Arg+4;

		break;

	case LDRM16:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u16Wrk1=pmMemory->read16(getRegister(u32Code & 0xff));

		setRegister(u32Wrk1,u16Wrk1);

		u32Addr=u32Arg+4;

		break;

	case LDRM32:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u32Wrk2=pmMemory->read32(getRegister(u32Code & 0xff));

		setRegister(u32Wrk1,u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case LDMR8:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u8Wrk1=(UI8)(getRegister(u32Code & 0xff) & 0x00ff);

		pmMemory->write8(getRegister(u32Wrk1),u8Wrk1);

		u32Addr=u32Arg+4;

		break;

	case LDMR16:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u16Wrk1=(UI16)(getRegister(u32Code & 0xff) & 0xffff);

		pmMemory->write16(getRegister(u32Wrk1),u16Wrk1);

		u32Addr=u32Arg+4;

		break;

	case LDMR32:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u32Wrk2=getRegister(u32Code & 0xff);

		pmMemory->write32(getRegister(u32Wrk1),u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case LDMI8:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u8Wrk2=(UI8)(u32Code & 0xff);

		pmMemory->write8(getRegister(u32Wrk1),u8Wrk2);

		u32Addr=u32Arg+4;

		break;

	case LDMI16:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u16Wrk2=pmMemory->read16(u32Arg+4);

		pmMemory->write16(getRegister(u32Wrk1),u16Wrk2);

		u32Addr=u32Arg+6;

		break;

	case LDMI32:

		u32Wrk1=(u32Code & 0xff00) >> 8;
		u32Wrk2=pmMemory->read32(u32Arg+4);

		pmMemory->write32(getRegister(u32Wrk1),u32Wrk2);

		u32Addr=u32Arg+8;

		break;

	case EXC:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		setRegister((u32Code & 0xff00) >> 8,getRegister(u32Code & 0xff));
		setRegister(u32Code & 0xff,u32Wrk1);

		u32Addr=u32Arg+4;

		break;

	case ADDRR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=getRegister(u32Code & 0xff);

		setSROF(u32OF(u32Wrk1,u32Wrk2));

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1+u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case ADDRI8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=u32Code & 0xff;

		setSROF(u32OF(u32Wrk1,u32Wrk2));

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1+u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case ADDRI32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(u32Arg+4);

		setSROF(u32OF(u32Wrk1,u32Wrk2));

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1+u32Wrk2);

		u32Addr=u32Arg+8;

		break;

	case ADDRM8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read8(getRegister(u32Code & 0xff));

		setSROF(u32OF(u32Wrk1,u32Wrk2));

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1+u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case ADDRM16:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read16(getRegister(u32Code & 0xff));

		setSROF(u32OF(u32Wrk1,u32Wrk2));

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1+u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case ADDRM32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(getRegister(u32Code & 0xff));

		setSROF(u32OF(u32Wrk1,u32Wrk2));

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1+u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case ADDMR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(u32Wrk1);
		u32Wrk3=getRegister(u32Code & 0xff);

		setSROF(u32OF(u32Wrk2,u32Wrk3));

		pmMemory->write32(u32Wrk1,u32Wrk2+u32Wrk3);

		u32Addr=u32Arg+4;

		break;

	case SUBRR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=getRegister(u32Code & 0xff);

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1-u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case SUBRI8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=u32Code & 0xff;

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1-u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case SUBRI32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(u32Arg+4);

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1-u32Wrk2);

		u32Addr=u32Arg+8;

		break;

	case SUBRM8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read8(getRegister(u32Code & 0xff));

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1-u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case SUBRM16:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read16(getRegister(u32Code & 0xff));

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1-u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case SUBRM32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(getRegister(u32Code & 0xff));

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		setRegister((u32Code & 0xff00) >> 8,u32Wrk1-u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case SUBMR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(u32Wrk1);
		u32Wrk3=getRegister(u32Code & 0xff);

		setSRSig(u32Wrk2<u32Wrk3);
		setSRZero(u32Wrk2==u32Wrk3);

		pmMemory->write32(u32Wrk1,u32Wrk2-u32Wrk3);

		u32Addr=u32Arg+4;

		break;

	case MULRR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=getRegister(u32Code & 0xff);

		if (u32Wrk2>0)
			setSROF(u32MulOF(u32Wrk1,u32Wrk2));
		else
			setSROF(false);

		u32Wrk=u32Wrk1*u32Wrk2;

		setRegister((u32Code & 0xff00) >> 8,u32Wrk);
		setSRZero(u32Wrk==0);

		u32Addr=u32Arg+4;

		break;

	case MULRI8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=u32Code & 0xff;

		if (u32Wrk2>0)
			setSROF(u32MulOF(u32Wrk1,u32Wrk2));
		else
			setSROF(false);

		u32Wrk=u32Wrk1*u32Wrk2;

		setRegister((u32Code & 0xff00) >> 8,u32Wrk);
		setSRZero(u32Wrk==0);

		u32Addr=u32Arg+4;

		break;

	case MULRI32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(u32Arg+4);

		if (u32Wrk2>0)
			setSROF(u32MulOF(u32Wrk1,u32Wrk2));
		else
			setSROF(false);

		u32Wrk=u32Wrk1*u32Wrk2;

		setRegister((u32Code & 0xff00) >> 8,u32Wrk);
		setSRZero(u32Wrk==0);

		u32Addr=u32Arg+8;

		break;

	case MULRM8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read8(getRegister(u32Code & 0xff));

		if (u32Wrk2>0)
			setSROF(u32MulOF(u32Wrk1,u32Wrk2));
		else
			setSROF(false);

		u32Wrk=u32Wrk1*u32Wrk2;

		setRegister((u32Code & 0xff00) >> 8,u32Wrk);
		setSRZero(u32Wrk==0);

		u32Addr=u32Arg+4;

		break;

	case MULRM16:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read16(getRegister(u32Code & 0xff));

		if (u32Wrk2>0)
			setSROF(u32MulOF(u32Wrk1,u32Wrk2));
		else
			setSROF(false);

		u32Wrk=u32Wrk1*u32Wrk2;

		setRegister((u32Code & 0xff00) >> 8,u32Wrk);
		setSRZero(u32Wrk==0);

		u32Addr=u32Arg+4;

		break;

	case MULRM32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(getRegister(u32Code & 0xff));

		if (u32Wrk2>0)
			setSROF(u32MulOF(u32Wrk1,u32Wrk2));
		else
			setSROF(false);

		u32Wrk=u32Wrk1*u32Wrk2;

		setRegister((u32Code & 0xff00) >> 8,u32Wrk);
		setSRZero(u32Wrk==0);

		u32Addr=u32Arg+4;

		break;

	case MULMR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(u32Wrk1);
		u32Wrk3=getRegister(u32Code & 0xff);

		if (u32Wrk3>0)
			setSROF(u32MulOF(u32Wrk2,u32Wrk3));
		else
			setSROF(false);

		u32Wrk=u32Wrk2*u32Wrk3;

		pmMemory->write32(u32Wrk1,u32Wrk);
		setSRZero(u32Wrk==0);

		u32Addr=u32Arg+4;

		break;

	case CPRR8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=getRegister((u32Code & 0xff) & 0xff);

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case CPRR16:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=getRegister((u32Code & 0xff) & 0xffff);

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case CPRR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=getRegister(u32Code & 0xff);

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case CPRI8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=u32Code & 0xff;

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case CPRI32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(u32Arg+4);

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+8;

		break;

	case CPRM8:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read8(getRegister(u32Code & 0xff));

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case CPRM16:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read16(getRegister(u32Code & 0xff));

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case CPRM32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=pmMemory->read32(getRegister(u32Code & 0xff));

		setSRSig(u32Wrk1<u32Wrk2);
		setSRZero(u32Wrk1==u32Wrk2);

		u32Addr=u32Arg+4;

		break;

	case JUMPI32:

		u32Addr=pmMemory->read32(u32Arg+4);

		break;

	case JUMPR32:

		u32Addr=getRegister((u32Code & 0xff00) >> 8);

		break;

	case JUMPRR32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
		u32Wrk2=getRegister(u32Code & 0xff);

		u32Addr=u32Wrk1+u32Wrk2;

		break;

	case JUMPM32:

		u32Wrk1=getRegister((u32Code & 0xff00) >> 8);

		u32Addr=pmMemory->read32(u32Wrk1);

		break;

	case JPZEI32:

		if (getSRZero())
			u32Addr=pmMemory->read32(u32Arg+4);
		else
			u32Addr=u32Arg+8;

		break;

	case JPZER32:

		if (getSRZero())
			u32Addr=getRegister((u32Code & 0xff00) >> 8);
		else
			u32Addr=u32Arg+4;

		break;

	case JPZERR32:

		if (getSRZero()) {

			u32Wrk1=getRegister((u32Code & 0xff00) >> 8);
			u32Wrk2=getRegister(u32Code & 0xff);

			u32Addr=u32Wrk1+u32Wrk2;

		} else
			u32Addr=u32Arg+4;

		break;

	case JPZEM32:

		if (getSRZero()) {

			u32Wrk1=getRegister((u32Code & 0xff00) >> 8);

			u32Addr=pmMemory->read32(u32Wrk1);

		} else
			u32Addr=u32Arg+4;

		break;

	default:

		throw invalid_argument("ȃR[h");

	}

	setPC(u32Addr);

	return true;

}

CCpu::~CCpu() { }
